
/* 456789a123456789b123456789c123456789d123456789e123456789f123456789g1234567*/

/*
 * lwf - print a text file on the LaserWriter
 *
 * Usage: lwf [-d] [-h] [-i#] [-l] [-olist] [-p] [-s#] [-t#] [-v] [file ...]
 *
 * Options:
 *	-d		Debug mode
 *	-h		Don't print header page
 *	-i#		Indent each line
 *	-l		Landscape instead of PORTRAIT
 *	-olist		Only print pages in the specified range
 *	-p		Use pr to print
 *	-s#		Use point size #
 *	-t#		Use tab size #
 *	-v		Verbose
 *
 * If no files are specified, stdin is used.
 * Page reversal is performed.
 * Form feeds should work
 * Backspacing with underlining (or overprinting) should work as expected
 *
 * Problems:
 *	- Should have:
 *		[-#n] option, where n=number of copies of each file or stdin
 *		[-T filename] option, where filename is used for title instead
 *		of a blank name
 *	- User can't give pr header if stdin is used
 *	- The volumne of PostScript output produced by this program could be
 *	  reduced somewhat (there may also be do-nothing sequences produced)
 *
 * BJB - Nov/85
 * ========================================================================
 *
 * Copyright (c) Barry Brachman
 *
 * Permission is given to freely copy or distribute this program
 * with the following restrictions:
 *
 *	1) The author is not responsible for the consequences of use of
 *		this software and is not responsible for correcting any defects
 *	2) The origin of this software must not be misrepresented, either by
 *		explicit claim or by omission
 *	3) You may not sell this program
 *	4) Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software
 *
 *
 * Barry Brachman
 * Dept. of Computer Science
 * Univ. of British Columbia
 * Vancouver, B.C. V6T 1W5
 *
 * .. {ihnp4!alberta, uw-beaver}!ubc-vision!ubc-cs!brachman
 * brachman@cs.ubc.cdn
 * brachman%ubc.csnet@csnet-relay.arpa
 * brachman@ubc.csnet
 * ========================================================================
 */

#include <sys/file.h>
#include <pwd.h>
#include <stdio.h>

#define PRCMD			"/bin/pr"	/* where pr is */

#define MAXPAGES		10000	/* maximum number of pages */
#define PORTRAIT_START_Y	768	/* first row (Y coord) on each page */
#define LANDSCAPE_START_Y	576
#define START_X			25	/* position of start of each line */
#define START_Y_HEADER		600	/* first row (Y coord) of header */

#define MAX_X			612
#define MAX_Y			792


#define NPSIZES			6
struct psize {
	int size;
	int portrait_page_length;
	int landscape_page_length;
} psize[NPSIZES] = {
	 7, 108, 80,
	 8,  94, 70,
	 9,  84, 62,
	10,  75, 56,
	11,  68, 51,
	12,  62, 46
};

#define USAGE		"[-d] [-h] [-i#] [-l] [-olist] [-p] [-s#] [-t#] [-v] [file ...]"

long page_map[MAXPAGES];		/* offset of first byte of each page */
int page_count;

int lines_per_page;
int point_size;
int tab_size;
int start_x, start_y;

char bufin[BUFSIZ];
char bufout[10240];			/* should be plenty big enough */
char bigbuf[32768];			/* max page size */

int currentp, row;
char *range;

int dflag, hflag, lflag, pflag, vflag;

char *fgetline();

char *progname;

main(argc, argv)
int argc;
char **argv;
{
	register int i, j, first_file;
	char *pc;
	struct psize *p, *get_psize();
	double offset, atof();
	char *rindex();
	FILE *infile, *popen();

	if ((pc = rindex(argv[0], '/')) != NULL)
		progname = pc + 1;
	else
		progname = argv[0];
	range = ":";
	hflag = 1;
	start_x = START_X;
	p = get_psize(10);
	tab_size = 8;
	for (i = 1; i < argc && argv[i][0] == '-'; i++) {
		switch (argv[i][1]) {
		case 'd':
			dflag = 1;
			break;
		case 'h':
			hflag = 0;
			break;
		case 'i':
			offset = atof(&argv[i][2]);
			if (offset < 0.0 || offset >= 8.5) {
				fprintf(stderr, "%s: bad indent\n", progname);
				exit(1);
			}
			start_x = (int) (offset * 72);
			break;
		case 'l':
			lflag = 1;
			break;
		case 'o':
			range = &argv[i][2];
			if (checkrange(range)) {
				fprintf(stderr, "%s: bad range specification\n",
								progname);
				exit(1);
			}
			break;
		case 'p':
			pflag = 1;
			break;
		case 's':
			j = atoi(&argv[i][2]);
			if ((p = get_psize(j)) == (struct psize *) NULL) {
				fprintf(stderr, "%s: bad point size\n", progname);
				exit(1);
			}
			break;
		case 't':
			tab_size = atoi(&argv[i][2]);
			break;
		case 'v':
			vflag = 1;
			break;
		default:
			fprintf(stderr, "Usage: %s %s\n", progname, USAGE);
			exit(1);
		}
	}

	for (j = i; j < argc; j++) {
		if (access(argv[j], R_OK) == -1) {
			fprintf(stderr, "%s: cannot access %s\n", progname,
								argv[j]);
			exit(1);
		}
	}

	if (lflag) {
		start_y = LANDSCAPE_START_Y;
		lines_per_page = p->landscape_page_length;
	}
	else {
		start_y = PORTRAIT_START_Y;
		lines_per_page = p->portrait_page_length;
	}
	point_size = p->size;

	preamble();

	first_file = i;
	if (i == argc) {	/* no files on command line */
		infile = stdin;
		if (pflag) {
			build_prcmd(bufin, "");
			if ((infile = popen(bufin, "r")) == (FILE *) NULL) {
				fprintf(stderr, "%s: can't popen\n", progname);
				exit(1);
			}
		}
		if (vflag)
			fprintf(stderr, "printing stdin\n");
		print(infile);
		if (pflag)
			pclose(infile);
	}

	for (j = argc - 1; i < argc; j--) {
		infile = stdin;
		if (pflag) {
			build_prcmd(bufin, argv[j]);
			if ((infile = popen(bufin, "r")) == (FILE *) NULL) {
				fprintf(stderr, "%s: can't popen\n", progname);
				exit(1);
			}
		}
		else {
			if (freopen(argv[j], "r", stdin) == NULL) {
				fprintf(stderr, "%s: can't open %s\n",
							progname, argv[j]);
				exit(1);
			}
		}
		if (vflag)
			fprintf(stderr, "printing %s\n", argv[j]);
		print(infile);
		if (pflag)
			pclose(infile);
		i++;
	}

	if (hflag)
		header(argc - first_file, argv + first_file);
	printf("restore\n");
}

/*
 * Return a pointer to the point size structure for the
 * specified point size
 */
struct psize *
get_psize(size)
int size;
{
	register int i;

	for (i = 0; i < NPSIZES; i++)
		if (psize[i].size == size)
			break;
	if (i == NPSIZES)
		return((struct psize *) NULL);
	return(&psize[i]);
}

/*
 * Initial lines sent to the LaserWriter
 */
preamble()
{
	printf("%%!\nsave\n");
	printf("initgraphics\n");
	printf("/Courier findfont %d scalefont\nsetfont\n", point_size);
	printf("/bkup { stringwidth pop neg 0 rmoveto } def\n");
	printf("/tab1 {( ) show} def\n");
	printf("/tab2 {(  ) show} def\n");
	printf("/tab3 {(   ) show} def\n");
	printf("/tab4 {(    ) show} def\n");
	printf("/tab5 {(     ) show} def\n");
	printf("/tab6 {(      ) show} def\n");
	printf("/tab7 {(       ) show} def\n");
	printf("/tab8 {(        ) show} def\n");
	if (lflag)
		printf("612 0 translate\n90 rotate\n");
	fflush(stdout);
}

/*
 * Generate a command, in the specified buffer, to print the given file
 * according to the options in effect
 */
build_prcmd(buf, file)
char *buf, *file;
{
	sprintf(buf, "%s -l%d %s", PRCMD, lines_per_page, file);
	if (vflag)
		fprintf(stderr, "pr cmd: %s\n", buf);
}

/*
 * Print a file
 *
 * The input stream may be stdin, a file, or a pipe
 * The output goes to a temporary file, and then reverse is called
 * to do the page reversal to stdout
 */
print(infile)
FILE *infile;
{
	register int eof, pagenum, r;
	register char *p;
	FILE *outfile;
	char *gets(), *mktemp();

	page_count = 0;
	page_map[0] = 0L;
	sprintf(bufin, "/tmp/%sXXXXXX", progname);
	if (vflag)
		fprintf(stderr, "temp will be: %s ... ", bufin);
	p = mktemp(bufin);
	if (vflag)
		fprintf(stderr, "%s\n", p);
	if ((outfile = fopen(p, "w+")) == (FILE *) NULL) {
		fprintf(stderr, "%s: can't create %s\n", progname, p);
		cleanup();
		/*NOTREACHED*/
	}
	if (!dflag)
		unlink(p);
	else
		fprintf(stderr, "will not unlink %s\n", p);
	pagenum = 1;
	eof = 0;
	while (!eof) {
		row = start_y;
		if ((r = inrange(pagenum, range)) == -1) {
			cleanup();
			/*NOTREACHED*/
		}
		else if (r == 1)
			eof = printpage(infile, outfile);
		else if (r == 0)
			eof = flushpage(infile);
		else {
			fprintf(stderr, "%s: bad inrange result\n", progname);
			cleanup();
			/*NOTREACHED*/
		}
		pagenum++;
	}
	if (row != start_y)
		endpage(outfile);
	fflush(outfile);
	reverse(outfile);
	fclose(outfile);
}

/*
 * Process the next page
 */
printpage(infile, outfile)
FILE *infile, *outfile;
{
	register int lineno;

	for (lineno = 0; lineno < lines_per_page; lineno++) {
		if (fgetline(bufin, sizeof(bufin), infile) == (char *) NULL)
			return(1);
		if (bufin[0] == '\f')
			break;
		if (bufin[0] != '\n') {
			fprintf(outfile, "%d %d moveto\n", start_x, row);
			proc(bufin, outfile);
			fprintf(outfile, "(%s) show\n", bufout);
		}
		row -= point_size;
	}
	endpage(outfile);
	return(0);
}

/*
 * The next page will not be printed; just consume the input and discard
 * Don't change xrow since we don't want an endpage()
 */
flushpage(infile)
FILE *infile;
{
	register int lineno, xrow;

	xrow = row;
	for (lineno = 0; lineno < lines_per_page; lineno++) {
		if (fgetline(bufin, sizeof(bufin), infile) == (char *) NULL)
			return(1);
		if (bufin[0] == '\f')
			break;
		xrow -= point_size;
	}
	return(0);
}

/*
 * A page has been written to the temp file
 * Record the start of the next page
 */
endpage(outfile)
FILE *outfile;
{
	long ftell();

	if (page_count == MAXPAGES) {
		fprintf(stderr, "%s: pagelimit (%d) reached\n",
						progname, MAXPAGES);
		cleanup();
		/*NOTREACHED*/
	}
	fprintf(outfile, "showpage\n");
	fflush(outfile);
	page_map[++page_count] = ftell(outfile);
	if (vflag)
		fprintf(stderr, "x");
}

/*
 * Print the pages to stdout in reverse order
 */
reverse(outfile)
FILE *outfile;
{
	register int fd, i, nbytes;

	if (vflag)
		fprintf(stderr, "\nreversing %d page%s\n", page_count,
						page_count > 1 ? "s" : "");
	fd = fileno(outfile);
	if (dflag) {
		for (i = 0; i <= page_count; i++)
			fprintf(stderr, "[%ld]\n", page_map[i]);
	}
	for (i = page_count - 1; i >= 0; i--) {
		if (lseek(fd, page_map[i], 0) == -1) {
			fprintf(stderr, "%s: seek error\n", progname);
			cleanup();
			/*NOTREACHED*/
		}
		nbytes = page_map[i + 1] - page_map[i];
		if (nbytes > sizeof(bigbuf)) {
			fprintf(stderr, "%s: page too big\n", progname);
			fprintf(stderr, "nbytes = %d\n", nbytes);
			cleanup();
			/*NOTREACHED*/
		}
		nbytes = read(fd, bigbuf, nbytes);
		write(1, bigbuf, nbytes);
	}
}

/*
 * Process a line, escaping characters when necessary and handling tabs
 * We are assuming that a line ends with a newline (fgets/fgetline)
 *
 * NOTE: lazy attempt at checking for buffer overflow on bufout
 */
proc(in, outfile)
char *in;
FILE *outfile;
{
	register int i;
	register char *last, *p, *q, *savep;

	currentp = 0;
	last = bufout + sizeof(bufout);
	for (p = in, q = bufout; *p != '\n' && *p != '\0' && q < last; p++) {
		switch (*p) {
		case '\r':
			*q++ = '\\';
			*q++ = 'r';
			currentp = 0;
			break;
		case '\t':
/*
			*q++ = '\\';
			*q++ = 't';
*/
			i = tab_size - (currentp % tab_size);
			sprintf(q, ") show tab%d (", i);
/*			sprintf(q, ") show (%*s) show (", i, ""); */
			q += strlen(q);
			currentp += i;
			break;
		case '\b':
/*
			*q++ = '\\';
			*q++ = 'b';
*/
			savep = p++;
			for (i = 1; *p == '\b'; p++)
				i++;
			*q = '\0';
			fprintf(outfile, "(%s) show\n", bufout);
			fprintf(outfile, "(%.*s) bkup ", i, savep - i);
/*
			fprintf(outfile, "(%.*s) stringwidth ", i, savep - i);
			fprintf(outfile, "pop neg 0 rmoveto\n");
*/
			currentp -= i;
			q = bufout;
			p--;
			break;
		case '\f':
/*
			*q++ = '\\';
			*q++ = 'f';
*/
			*q = '\0';
			fprintf(outfile, "(%s) show\n", bufout);
			endpage(outfile);
			row = start_y;
			fprintf(outfile, "%d %d moveto\n", start_x, row);
			q = bufout;
			currentp = 0;
			break;
		case '\\':
			*q++ = '\\';
			*q++ = '\\';
			currentp++;
			break;
		case '(':
			*q++ = '\\';
			*q++ = '(';
			currentp++;
			break;
		case ')':
			*q++ = '\\';
			*q++ = ')';
			currentp++;
			break;
		default:
			*q++ = *p;
			currentp++;
			break;
		}
	}
	*q = '\0';
	if (q >= last) {
		fprintf(stderr, "%s: bufout overflow\n", progname);
		cleanup();
		/*NOTREACHED*/
	}
}

/*
 * Print a header page
 */
header(nfiles, files)
int nfiles;
char **files;
{
	register int i;
	register char *p;
	long t, time();
	char *ctime(), *gethostname(), *getlogin();
	struct passwd *pw, *getpwuid();

	if (vflag) {
		fprintf(stderr, "printing header\n");
		fprintf(stderr, "%d file%s are:\n", nfiles,
							nfiles > 1 ? "s" : "");
		if (nfiles == 0)
			fprintf(stderr, "\tstdin\n");
		for (i = 0; i < nfiles; i++)
			fprintf(stderr, "\t%s\n", files[i]);
	}
	printf("initgraphics\n");
	printf("/Times-Roman findfont 18 scalefont\nsetfont\n");
	row = START_Y_HEADER;
	printf("newpath %d %d moveto\n", START_X, row);
	printf("%d %d lineto\n", START_X + 400, row);
	row -= 6;
	printf("%d %d moveto\n", START_X, row);
	printf("%d %d lineto\n", START_X + 400, row);
	row -= 24;
	printf("2 setlinewidth\n");
	printf("stroke\n");

	if ((p = getlogin()) == (char *) NULL) {
		if ((pw = getpwuid(getuid())) == NULL)
			p = "Whoknows";
		else
			p = pw->pw_name;
		endpwent();
	}
	printf("%d %d moveto\n", START_X, row);
	printf("(User: %s) show\n", p);
	row -= 24;
	gethostname(bufin, sizeof(bufin));
	printf("%d %d moveto\n", START_X, row);
	printf("(Host: %s) show\n", bufin);
	row -= 24;
	t = time(0);
	p = ctime(&t);
	p[strlen(p) - 1] = '\0';
	printf("%d %d moveto\n", START_X, row);
	printf("(Date: %s) show\n", p);
	row -= 24;

	if (nfiles == 0) {
		printf("%d %d moveto\n", START_X, row);
		printf("(File: stdin) show\n", bufout);
	}
	else {
		printf("%d %d moveto\n", START_X, row);
		for (i = 0, p = bufin; i < nfiles; i++) {
			sprintf(p, "%s%s", files[i],
					   i != nfiles - 1 ? ", " : "");
			p += strlen(p);
		}
		proc(bufin, stdout);
		printf("(File%s: %s) show\n", nfiles > 1 ? "s" : "", bufout);
	}

	row -= 12;
	printf("newpath %d %d moveto\n", START_X, row);
	printf("%d %d lineto\n", START_X + 400, row);
	row -= 6;
	printf("%d %d moveto\n", START_X, row);
	printf("%d %d lineto\n", START_X + 400, row);
	printf("stroke\n");
	printf("showpage\n");
}

/*
 * Special version of fgets
 * Read until a formfeed or a newline
 * If a formfeed is the first character, return it immediately
 * If a formfeed is found after the first character, replace it by a newline
 * and push the formfeed back onto the input stream
 * A special case is a formfeed followed by a newline in which case the
 * newline is ignored 
 */
char *
fgetline(s, n, iop)
char *s;
int n;
register FILE *iop;
{
	register int ch;
	register char *cs;

	cs = s;
	if (--n > 0) {
		if ((ch = getc(iop)) == EOF)
			return(NULL);
		if (ch == '\f') {
			if ((ch = getc(iop)) != '\n')
				if (ungetc(ch, iop) == EOF)
					fprintf(stderr, "fgetline - ungetc\n");
			strncpy(cs, "\f\n", n);
			return(s);
		}
	}
	while (n-- > 0) {
		if (ch == '\f')
			break;
		*cs++ = ch;
		if (ch == '\n')
			break;
		if ((ch = getc(iop)) == EOF)
			break;
	}

	if (ch == EOF && cs == s)
		return(NULL);
	if (ch == '\f') {
		*cs++ = '\n';
		if (ungetc(ch, iop) == EOF)
			fprintf(stderr, "fgetline can't ungetc\n");
	}
	*cs++ = '\0';
	return(s);
}

/*
 * Clean up and exit after an error
 */
cleanup()
{

	exit(1);
}
